import sys 
import os
from pathlib import Path 
cwd = str(os.getcwd())
parent_dir = str(Path(os.getcwd()).parent)
sys.path.append(f'{parent_dir}/short_text_tagger/') # if testing from within tests/
sys.path.append(f'{cwd}/short_text_tagger/') # if testing from parent directory
import pandas as pd 
import pytest 
from edgelist import EdgeList
from nsbm import NSBM

corpus1 = pd.Series([
    ["store","love","hair","products"],
    [],
    ["communication","key","1"]
])
e1 = EdgeList(corpus1,weighted=True,directed=False)

def test_graph_and_nsbm_creation_naive():
    nsbm = NSBM(e1,block_level = 2)
    assert nsbm.g is not None 
    assert len(nsbm.g.get_vertices()) == 7, "the graph doesn't have the correct number of vertices"
    assert len(nsbm.g.get_edges()) == 9, "the graph doesn't have the correct number of edges"
    nsbm.fit()
    assert nsbm.state is not None, "the NSBM state needs to exist after fit() is called"


import sys 
import os
from pathlib import Path 
cwd = str(os.getcwd())
parent_dir = str(Path(os.getcwd()).parent)
sys.path.append(f'{parent_dir}/short_text_tagger/') # if testing from within tests/
sys.path.append(f'{cwd}/short_text_tagger/') # if testing from parent directory
print(list(sys.path))
from short_text_tagger.short_text_tagger import generate_topic_distributions_from_corpus
import pytest 
import pandas as pd


def test_smoke():

    short_texts_df = pd.read_csv("data/input.csv")
    df = generate_topic_distributions_from_corpus(short_texts_df)
    print(df)
    assert len(df) == len(short_texts_df)
test_smoke()


import sys 
import os
from pathlib import Path 
cwd = str(os.getcwd())
parent_dir = str(Path(os.getcwd()).parent)
sys.path.append(f'{parent_dir}/short_text_tagger/') # if testing from within tests/
sys.path.append(f'{cwd}/short_text_tagger/') # if testing from parent directory

from short_text_tagger.edgelist import EdgeList
import pytest 
import pandas as pd


corpus1 = pd.Series([
    ["store","love","hair","products"],
    [],
    ["communication","key","1"]
])
corpus2 = pd.Series([],dtype=str)


def test_raw_directed_edgelist():

    e1 = EdgeList(corpus1,weighted=False,directed=True)
    e2 = EdgeList(corpus2,weighted=False,directed=True)

    assert e1 is not None, "edge list needs to exist"
    assert e2 is not None, "edge list needs to exist" 

    assert e2.edgelist.equals(pd.DataFrame({"source":[],"target":[]})), "edge lists do not match"
    assert e1.edgelist.equals(pd.DataFrame({
        "source":["store","store","store","love","love","hair","communication","communication","key"],
        "target":["love","hair","products","hair","products","products","key","1","1",]
    })), "edge lists do not match"


def test_raw_undirected_edgelist():

    e1 = EdgeList(corpus1,weighted=False,directed=False)
    e2 = EdgeList(corpus2,weighted=False,directed=False)

    assert e1 is not None, "edge list needs to exist"
    assert e2 is not None, "edge list needs to exist" 

    assert e2.edgelist.equals(pd.DataFrame({"source":[],"target":[]})), "edge lists do not match"
    assert e1.edgelist.equals(pd.DataFrame({
        "source":["hair","hair","hair","love","love","products","1","1","communication"],
        "target":["love","products","store","products","store","store","communication","key","key",]
    })), "edge lists do not match"





from nltk.corpus import stopwords
import numpy as np


stop_words: set = set(stopwords.words('english'))


def string_to_valid_word_list(string:str) -> list:
    """ creates list of lowercase non-stopwords with no special characters from input string"""

    if not isinstance(string,str) or string == "":
        return []
    else:
        string = lower_case(string)
        string = remove_special_characters(string)
        return split_string(string)


def lower_case(string:str) -> str:
    """ lower case all words """
    assert isinstance(string,str), "cannot lowercase non-string datatype"
    return string.lower()


def remove_special_characters(string:str) -> str:
    """ removing special characters from string """
    assert isinstance(string,str), "cannot remove special characters from non-string datatype"
    string_with_valid_characters = ""
    for character in string:
        if character.isalnum() or character == ' ':
            string_with_valid_characters += character
    return string_with_valid_characters


def split_string(string:str) -> list:
    """ splits string into list of words, removing empty strings and stopwords """
    assert isinstance(string,str), "cannot split non-string datatype"
    word_list = string.split(" ")
    return [word for word in word_list if (word != "" and not word in stop_words)]





from itertools import combinations, chain
import pandas as pd


class EdgeList:

    def __init__(self, corpus: pd.Series, directed: bool = False, weighted: bool = True):
        
        self.weighted = weighted
        self.directed = directed
        self.corpus: pd.Series = corpus # item: list of words
        self.edgelist: pd.DataFrame = self.generate_edgelist()


    def generate_edgelist(self) -> pd.DataFrame:
        """ provided series of word lists, generate (weighted) edge list"""

        print("Generating edgelist")
        raw_edgelist = self.generate_raw_edgelist()
        if self.weighted:
            weighted_edgelist = self.weight_edgelist(raw_edgelist)
            return weighted_edgelist
        else:
            return raw_edgelist


    def weight_edgelist(self,raw_edgelist) -> pd.DataFrame:
        """ weights raw edgelist """

        raw_edgelist['weight'] = 1
        weighted_edgelist = raw_edgelist.groupby(['source','target']).agg({'weight':'count'}).reset_index()
        return weighted_edgelist 


    def generate_raw_edgelist(self) -> pd.DataFrame:
        """ provided series of word lists, generate unweighted edge list (duplicate edges allowed) """

        if len(self.corpus) == 0:
            # insert warning here
            return pd.DataFrame({"source":[],"target":[]})

        if not self.directed:
            words_combinations_generator = chain.from_iterable(combinations(sorted(words),2) for i, words in self.corpus.iteritems() if len(words)>1)
        else:
            words_combinations_generator = chain.from_iterable(combinations(words,2) for i, words in self.corpus.iteritems() if len(words)>1)
        source,target = zip(*words_combinations_generator)
        raw_edgelist = pd.DataFrame({
                                    'source': list(source),
                                    'target':list(target)
                                    })
        return raw_edgelist




from edgelist import EdgeList
import graph_tool.all as gt
import pandas as pd
from collections import defaultdict

class NSBM:

    def __init__(self, edgelist: EdgeList, block_level: int = 2):
        # assigning node names to unique integer identifiers
        node_names = edgelist.edgelist[["source","target"]].stack().reset_index(name="node_name")
        unique_node_names = node_names["node_name"].unique()

        # edgelist metadata
        self.node_name_to_node_index = {unique_node_names[i]:i for i in range(len(unique_node_names))}
        self.node_index_to_node_name = {index:name for name,index in self.node_name_to_node_index.items()}
        self.directed = edgelist.directed
        self.edgelist = edgelist.edgelist
        self.edgelist['source_index'] = self.edgelist['source'].apply(lambda name:self.node_name_to_node_index[name])
        self.edgelist['target_index'] = self.edgelist['target'].apply(lambda name:self.node_name_to_node_index[name])

        # graph
        self.g = self.__graph_from_edgelist() 
        
        # NSBM
        self.state = None 
        self.levels = None 
        self.block_level = block_level
        self.block_index_to_vertex_list_dict = defaultdict(list)
        self.block_index_to_block_name_dict = dict()
    


    def __graph_from_edgelist(self) -> gt.Graph:
        """ given edgelist, generate graph-tool graph """ 

        print("Generating graph from edgelist")
        g = gt.Graph(directed = self.directed)

        # initialize edge property
        weight_prop = g.new_edge_property("int")
        g.edge_properties['weight'] = weight_prop

        # initialize node property
        node_name_property = g.new_vertex_property("string")
        g.vertex_properties['node_name'] = node_name_property

        # create graph from edgelist
        g.add_edge_list(edge_list = self.edgelist[['source_index','target_index']].values,
                        eprops = [weight_prop])

        print("Number of vertices: ",len(g.get_vertices()))
        print("Number of edges", len(g.get_edges()))
        
        # add names to each node
        for v in g.vertices():
            node_name_property[v] = self.node_index_to_node_name[int(v)]

        # add weight to each edge 
        for i,e in enumerate(g.edges()):
            weight_prop[e] = self.edgelist['weight'][i]
        print("Graph created")
        return g



    def fit(self) -> None:
        """ given graph, fits NSBM and writes to state and block instance variables """
        
        print("Fitting NSBM")
        state = gt.minimize_nested_blockmodel_dl(self.g, deg_corr=True)
        self.state = state
        self.levels = self.state.get_levels()
        self.block_level = min(self.block_level,len(self.levels))
        self.__get_block_metadata()

        print("NSBM hierarchy summary:")
        state.print_summary()



    def get_block_from_node(self, node:int) -> int:
        """ given node, returns block index at self.block_level """

        block = node
        for i in range(self.block_level):
            block = self.levels[i].get_blocks()[block]
        return block 


    def __get_block_metadata(self) -> None:
        """ writes to block_index_to_vertex_list_dict & block_index_to_block_name_dict"""

        for v in self.g.vertices():
            block = self.get_block_from_node(int(v))
            self.block_index_to_vertex_list_dict[block].append(v)
  
        for block, vertices in self.block_index_to_vertex_list_dict.items():
            highest_degree_names = []
            degree = 0
            for vertex in vertices:
                if vertex.out_degree() > degree:
                    degree = vertex.out_degree()
            for vertex in vertices:
                if vertex.out_degree() == degree:
                    highest_degree_names.append(self.node_index_to_node_name[int(vertex)])
            self.block_index_to_block_name_dict[block] = sorted(highest_degree_names)[0] # always guarantee same node



import nltk
import pandas as pd
from short_text_tagger.text import string_to_valid_word_list
from edgelist import EdgeList
from nsbm import NSBM
from collections import defaultdict
import sys

nltk.download('stopwords')


def cleaned_texts_df_from_data(short_texts_df:pd.DataFrame) -> pd.DataFrame:
    """ Returns cleaned short texts dataframe 

    Parameters
    ----------
    short_texts_df : pd.DataFrame
        "id" and "text" columns

    Returns
    -------
    short_texts_df : pd.DataFrame
        additional "words" column
    """

    assert 'id' in short_texts_df.columns, "'id' is a required column name for this analysis" 
    assert 'text' in short_texts_df.columns, "'text' is a required column name for this analysis" 
    assert len(short_texts_df) > 0, "Need at least one short text for this analysis"

    short_texts_df = short_texts_df[['id','text']]
    short_texts_df['words'] = short_texts_df['text'].apply(lambda text: string_to_valid_word_list(text)) 
    return short_texts_df



def word_to_block_dict(edgelist: EdgeList, block_level: int) -> dict:
    """ Returns a word to topic (block) map for every word in edgelist by fitting an NSBM 
        from the edgelist and assigning each node to its community

    Parameters
    ----------
    edgelist : EdgeList

    Returns
    -------
    dict [word:block] 
        
    """
    nsbm = NSBM(edgelist,block_level = block_level)
    nsbm.fit()
    return {word: nsbm.block_index_to_block_name_dict[nsbm.get_block_from_node(i)] 
            for word, i in nsbm.node_name_to_node_index.items()}



def assign_text_probabilities(short_texts_df: pd.DataFrame, word_to_block_dict_list: list) -> pd.DataFrame:
    """ Assigns topic probabilities to short_texts_df 

    Parameters
    ----------
    short_texts_df : pd.DataFrame
        "id", "text", and "words" columns
    word_to_block_dict_list: List[Dict]
        list of word to block mappings

    Returns
    -------
    short_texts_df : pd.DataFrame
        additional columns of form "topic_prob__<topic>"
    """

    # initializing dict of topic probabilities for each short text
    text_topic_probabilities = {row['id']: defaultdict(int) for i,row in short_texts_df.iterrows()}
    topics_seen = set()
    for word_to_block in word_to_block_dict_list:
        topics_seen = topics_seen.union(set(word_to_block.values()))

    # for each text, get probabilities of topic memberships
    for _,row in short_texts_df.iterrows():
        for word in row['words']:
            for word_to_block in word_to_block_dict_list:
                block = word_to_block[word]
                text_topic_probabilities[row['id']][block] += 1/(len(word_to_block_dict_list)*len(row['words']))

    for topic in topics_seen:
        short_texts_df[f'topic_prob__{topic}'] = short_texts_df['id'].apply(lambda id: text_topic_probabilities[id][topic] if topic in text_topic_probabilities[id].keys() else 0) 
  
    return short_texts_df


def generate_topic_distributions_from_corpus(short_texts_df:pd.DataFrame, 
                                             iterations:int = 5, 
                                             weighted:bool = True,
                                             directed:bool = False,
                                             block_level = 2) -> pd.DataFrame: 


    short_texts_df = cleaned_texts_df_from_data(short_texts_df)
    edgelist = EdgeList(corpus = short_texts_df['words'], directed = directed, weighted = weighted)
    word_to_block_dict_list = [word_to_block_dict(edgelist,block_level) for _ in range(iterations)]
    short_texts_df = assign_text_probabilities(short_texts_df,word_to_block_dict_list)
    return short_texts_df





import sys 
import os
from pathlib import Path 
cwd = str(os.getcwd())
parent_dir = str(Path(os.getcwd()).parent)
sys.path.append(f'{parent_dir}/short_text_tagger/') # if testing from within tests/
sys.path.append(f'{cwd}/short_text_tagger/') # if testing from parent directory

from short_text_tagger import text
from nltk.corpus import stopwords
import pytest

stop_words: set = set(stopwords.words('english'))

def test_lower_case(): 
    s1 = "abcdefgh"
    s2 = "ABCDEFGH"
    s3 = "12345"
    s4 = " "

    assert text.lower_case(s1) == s1 
    assert text.lower_case(s2) == s1
    assert text.lower_case(s3) == s3
    assert text.lower_case(s4) == s4

def test_remove_special_characters():
    s1 = "abcd$%^"
    s2 = "^&%$#~"
    s3 = " "
    s4 = "I went shopp&^%**ing"
    assert text.remove_special_characters(s1) == "abcd"
    assert text.remove_special_characters(s2) == ""
    assert text.remove_special_characters(s3) == " "
    assert text.remove_special_characters(s4) == "I went shopping"
     


def test_split_string():
    s1 = ""
    s2 = "Go find the treasure"
    s3 = "and but or "

    assert text.split_string(s1) == []
    assert text.split_string(s2) == ['Go', 'find', 'treasure']
    assert text.split_string(s3) == []


def test_string_to_valid_word_list():
    s1 = "I went to the store to buy coffee"
    s2 = ""
    s3 = "     "
    s4 = "Sentence with CAPITALIZATION and special %^&*%$ !!!"

    assert text.string_to_valid_word_list(s1) == ["went","store","buy","coffee"]
    assert text.string_to_valid_word_list(s2) ==  []
    assert text.string_to_valid_word_list(s3) == []
    assert text.string_to_valid_word_list(s4) == ["sentence","capitalization","special"]